#!/usr/bin/env python
import sys
from copy import deepcopy, copy

class DiffSet:
    def __init__(self, ps, l):
        self.ps = ps
        self.diff = []
        self.old = ''
        self.new = ''
        self.added = ''

        l = l.split(' ')
        self.old_loc, self.old_size = map(int, l[1].split(','))
        self.new_loc, self.new_size = map(int, l[2].split(','))
        self.old_loc = -self.old_loc

        self.parse()

    def parse(self):

        num_lines_expected = self.new_size 

        while num_lines_expected:
            l = self.ps.f.readline()
            self.diff.append(l)
            c = l[0]
            l = l[1:]
            if c == '-':
                num_lines_expected += 1
                self.old += l
            elif c == '+':
                self.new += l
                self.added += l
            else:
                self.old += l
                self.new += l
            num_lines_expected -= 1

    def __str__(self):
        res = "@@ -%d,%d +%d,%d @@\n" % \
                      (self.old_loc, self.old_size, self.new_loc, self.new_size)
        res += ''.join(self.diff)
        return res

class FakeDiffSet(DiffSet):
    def __init__(self, diff, old, new, added, old_loc, new_loc, old_size, new_size):
        self.diff = deepcopy(diff)
        self.old = copy(old)
        self.new = copy(new)
        self.added = copy(added)
        self.old_loc, self.old_size = old_loc, old_size
        self.new_loc, self.new_size = new_loc, new_size

class PatchFile:
    def __init__(self, ps, l, junk):
        self.ps = ps
        self.junkstart = junk 
        self.junkend = ''
        self.diff = []
        self.minusline = l
        self.plusline = self.ps.f.readline()
        self.name = self.minusline[4:].split('\t')[0]

    def parse(self):

        while 1:
            l = self.ps.f.readline()
            if not l:
                return
            if l.startswith('@@'): 
                ds = DiffSet(self.ps, l)
                self.diff.append(ds)
                continue
            elif self.diff: # we're in diff parsing mode, no junk expected
                return l
            self.junkstart += l

    def __str__(self):
        res = self.junkstart + self.minusline + self.plusline
        for d in self.diff:
            res += str(d)
        return res + self.junkend

class PatchSet:

    def __init__(self, fname):
        self.f = open(fname, "r")
        self.indices = {}
        self.patchorder = []
        self.junk = ''
        self.parse()

    def parse(self):
        pf = None
        junkcollect = ''
        already_read = False
        while 1:
            if not already_read:
                l = self.f.readline()
            else:
                already_read = False
            if not l:
                return
            if l.startswith('Index:'):
                junkline = l + self.f.readline()
                junkcollect += junkline
                minusline = self.f.readline()
                if not minusline.startswith('---'):
                    # whoops - this is why you use LALR parsing readahead...
                    already_read = True
                    l = minusline
                    continue

                pf = PatchFile(self, minusline, junkcollect)
                l = pf.parse()
                junkcollect = ''
                self.indices[pf.name] = pf
                self.patchorder.append(pf.name)
                if l:
                    # whoops - this is why you use LALR parsing readahead...
                    already_read = True
            else:
                if pf is None:
                    self.junk += l
                else:
                    junkcollect += l

    def __str__(self):
        res = self.junk
        for k in self.patchorder:
            res += str(self.indices[k])
        return res[:-1]

if __name__ == '__main__':
    p = PatchSet("./f")
    print str(p)
